package jitas.admin;

import java.io.FileNotFoundException;
import jitas.util.Settings;
import java.io.IOException;
import java.rmi.RemoteException;
import java.util.Set;
import java.util.SortedSet;
import java.util.TreeSet;

import javax.swing.table.TableModel;

import jess.JessException;

import jitas.core.Admin;
import jitas.core.Student;
import jitas.exception.AddUserDeniedException;
import jitas.exception.DomainNotFoundException;
import jitas.exception.DomainNotLoadedException;
import jitas.exception.IllegalUserNameException;
import jitas.exception.StudentModelAlreadyExistsException;
import jitas.exception.UserNotFoundException;
import jitas.exception.UserNotLoggedInException;

/**
 * This class acts as an intermediate pathway
 * between the gui (AdminFrame) and jitas.core.Admin
 * @author lsf21 & rle41
 *
 */
public class GUIHelper {

	String currentFilter = null;//is null if AllDomains is selected as filter
	String current = null;
	
	TableModelCreator creator;
	Admin admin;
	
	Set<String> unselectedDomainsForNewStudent;
	Set<String> selectedDomains;
	
	boolean usernameEntryStatus = false;// Username entered is valid
	boolean passwordEntryStatus = false;// password entered is valid
	
	public GUIHelper() throws IOException {
		admin = new Admin(Settings.getInstance().getRootPath());
		creator = new TableModelCreator(admin);
	}
	
	
	/**
	 * Used to return an array of student usernames depending on
	 * what the filter is set to.  In the gui, these usernames
	 * will make up the list of students.
	 * @return the array of students dependent on the filter
	 * @throws FileNotFoundException
	 */
	public String[] getStudentUsernameStrings() throws FileNotFoundException{
		admin.refreshStudentList();
		Set<String> usernames = new TreeSet<String>();
		if(currentFilter == null){
			usernames =  admin.getAllStudentUsernameStrings();
		} else {
			usernames = admin.getStudentUsernameStrings(currentFilter);
		}
		
		return usernames.toArray(new String[0]);
	}
	
	
	/**
	 * Sets the current student which is selected in the gui
	 * @param username
	 */
	public void setCurrentUser(String username) {
		current = username;
	}

	/**
	 * Called when the remove button is pressed in the gui.
	 * This method then gets the admin class to remove the domain
	 * from the current student.
	 * @param row - current domain selected in the gui
	 * @throws DomainNotFoundException
	 * @throws DomainNotLoadedException
	 * @throws UserNotFoundException
	 * @throws UserNotLoggedInException
	 * @throws RemoteException
	 * @throws JessException
	 */
	public void removeDomain(int row) throws DomainNotFoundException, DomainNotLoadedException, UserNotFoundException, UserNotLoggedInException, RemoteException, JessException {
		admin.removeUser(current, creator.getDomainNameFromRow(row));
	}
	
	
	/**
	 * Called when the add button is pressed in the gui.
	 * Adds the selected domain to the set of domains a student
	 * can access.
	 * @param row - the row of the domain selected
	 * @throws DomainNotFoundException
	 * @throws DomainNotLoadedException
	 * @throws StudentModelAlreadyExistsException
	 * @throws UserNotFoundException
	 * @throws RemoteException
	 * @throws JessException
	 * @throws AddUserDeniedException 
	 */
	public void addDomain(int row) throws DomainNotFoundException, DomainNotLoadedException, 
	StudentModelAlreadyExistsException, UserNotFoundException, RemoteException, JessException, AddUserDeniedException {
		admin.addAllowedDomain(current, creator.getDomainNameFromRow(row));
	}
	
	/**
	 * gets the set of all domain names
	 * @return a SortedSet of all domain names
	 */
	public SortedSet<String> getAllDomainNames() {
		return admin.getAllDomainNames();
	}
	
	
	/**
	 * gets an array of all domain names
	 * @return an array of all domain names
	 */
	public String[] getAllDomainNameStrings() {
		return admin.getAllDomainNames().toArray(new String[0]);
	}
	
	/**
	 * gets the table model for the current student
	 * @return a tableModel
	 */
	public TableModel getStudentTableModel() {
		return creator.getModel(current);
	}
	
	/**
	 * tests whether a student has access to a certain domain
	 * @param row - the row of the domain currently selected
	 * @return true if the student does have access to domain
	 * @throws UserNotFoundException
	 */
	public boolean isAllowedInDomain(int row) throws UserNotFoundException {		
		return admin.studentHasAccountInDomain(current, creator.getDomainNameFromRow(row));
	}

	/**
	 * Sets the domaain of which we wish to filter by
	 * If the filter is null, then all domains are selected
	 * @param filter - the domain name we wish to filter by or null if all domains are selected
	 */
	public void setFilter(String filter) {
		currentFilter = filter;
	}
	
	
	/**
	 * Resets the currently selected domains which 
	 * the new student will belong to, to the default
	 * which is: no domains are selected.
	 */
	public void resetNewStudentDomainLists() {
		unselectedDomainsForNewStudent = admin.getAllDomainNames();
		selectedDomains = new TreeSet<String>();
	}
	
	
	/**
	 * gets the domains which the new student will not currently
	 * belong to.
	 * @return an array of all the domains the student does not currently have access
	 */
	public String[] getUnselectedDomainStrings() {
		
		return unselectedDomainsForNewStudent.toArray(new String[0]);
	}
	
	
	/**
	 * gets all the domains names which are currently selected.
	 * This is used in creating a new student.
	 * @return an array of domain names
	 */
	public String[] getSelectedDomainStrings() {
		return selectedDomains.toArray(new String[0]);
	}
	
	
	/**
	 * checks whether the student with username already exists
	 * @param username -username of the student
	 * @return a message
	 */
	public String checkUsername(String username) {
		for (String s: admin.getAllStudentUsernameStrings()) {
			if (username.equals(s)) {
				usernameEntryStatus = false;
				return "Invalid Username: User already exists";
			}
		}
		
		
		int status = Student.checkUsername(username);
		switch (status) {
			case 0:
				usernameEntryStatus = true;
				return "Valid Username";
				
			case 1:	
				usernameEntryStatus = false;
				return "Invalid Username: Disallowed characters have been used.";
				
			case 2:
				usernameEntryStatus = false;
				return "Invalid Username: Must be at least 3 characters";	
		}
		return "This highly unexpected";
	}
	
	
	/** 
	 * Checks whether a password is valid or not, and returns a 
	 * message.
	 */
	public String checkPassword(String password) {
		int status = Student.checkPassword(password);
		switch (status) {
			case 0:
				passwordEntryStatus = true;
				return "Valid Password";
			case 2:
				passwordEntryStatus = false;
				return "Invalid Password: Must be at least 3 characters";
		}
		return "This highly unexpected";
	}
	
	
	/**
	 * checks whether the username, password are valid,
	 * and also whether there is at least one domain from calls to 
	 * addDomainToSelectedDomains.
	 * @return True if username, password and selection domain information is valid.
	 */
	public boolean allInputValid() {
		return usernameEntryStatus && passwordEntryStatus && selectedDomains.size()>0;
	}
	
	
	/**
	 * Creates a new student with all domains from previous calls to 
	 * addDomainToSelectedDomains.
	 * @param username - the students username
	 * @param password - the students password
	 * @throws AddUserDeniedException
	 * @throws DomainNotFoundException
	 * @throws DomainNotLoadedException
	 * @throws IllegalUserNameException
	 * @throws StudentModelAlreadyExistsException
	 * @throws RemoteException
	 * @throws JessException
	 */
	public void addNewStudent(String username, String password) throws 
	AddUserDeniedException, DomainNotFoundException, DomainNotLoadedException, 
	IllegalUserNameException, StudentModelAlreadyExistsException, 
	RemoteException, JessException {
		admin.addUser(username, password, selectedDomains);
	}
	
	
	/**
	 * Adds a domain to the set of domains a new student will have
	 * access to.
	 * @param domain - the domain name
	 */
	public void addDomainToSelectedDomains (String domain) {
		unselectedDomainsForNewStudent.remove(domain);
		selectedDomains.add(domain);
	}
	
	
	/**
	 * Removes a domain from the domains a new student will have access
	 * to.
	 * @param domain - the domain name
	 */
	public void removeDomainFromSelectedDomains (String domain) {
		selectedDomains.remove(domain);
		unselectedDomainsForNewStudent.add(domain);
	}
	
	
	/**
	 * @return true if the username entered is valid
	 */
	public boolean getUsernameEntryStatus() {
		return usernameEntryStatus;
	}
	
	/**
	 * @return true if the password entered is valid
	 */
	public boolean getPasswordEntryStatus() {
		return passwordEntryStatus;
	}
	
	
	/**
	 * deletes a student from the entire JITAS System
	 * @param username - the username of the student
	 * @throws DomainNotFoundException
	 * @throws DomainNotLoadedException
	 * @throws UserNotFoundException
	 * @throws UserNotLoggedInException
	 * @throws RemoteException
	 * @throws JessException
	 */
	public void deleteUser(String username) throws DomainNotFoundException, 
	DomainNotLoadedException, UserNotFoundException, UserNotLoggedInException, 
	RemoteException, JessException {
		admin.deleteUser(username);
	}
	
	
}
