package jitas.core;

import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;
import java.util.SortedMap;
import java.util.regex.Matcher;
import java.util.regex.Pattern;
import java.io.FileNotFoundException;
import java.io.UnsupportedEncodingException;
import java.security.NoSuchAlgorithmException;
import jitas.exception.*;
import jitas.util.*;
import jitas.xml.parser.StudentParser;
import jitas.xml.writer.StudentWriter;

/**
 * Student keeps track of all the domains a student is logged into
 * as well as the domains that the student is able to log into, and their
 * userName and password which the student uses to log into domains.
 * 
 * @author	DeathMarch (c314g2)
 * @version	08 October 2007
 */
public class Student {
	private String userName, password;	
	private Encryptor encryptor = null;
	private Set<String> allowedDomains;//Set of domains the Student can log into
	private Set<String> liveDomains = new HashSet<String>();
	private Map<String, StudentModel> models; //<domainName, StudentModel>
	private Map<String, String> currentSubdomains;

	/**
	 * Constructor.
	 * @param userName The students userName.
	 * @param password The encrypted password
	 * @param allowedDomains The domains a student can log into
	 * @throws NoSuchAlgorithmException 
	 */
	public Student(String userName, String password, Set<String> allowedDomains) throws IllegalUserNameException{
		Pattern p = Pattern.compile("[a-zA-Z_0-9]*");
		Matcher m = p.matcher(userName);
		if(!m.matches()){
			throw new IllegalUserNameException("Student.constructor: The following characters are not allowed in a " +
					"username: "+ userName + " has illegal characters, Only alphabet characters, " +
							"numbers and underscores allowed");
		}
		if (userName.length() < 3) {
			throw new IllegalUserNameException("Student.constructor: This is an invalid " +
					"username: \""+ userName + "\" is too short, usernames must contain at least 2 characters");
		}
		encryptor = Encryptor.getEncryptor();
		this.userName = userName;		
		this.password = password;
		this.allowedDomains = allowedDomains;	
		models = new HashMap<String, StudentModel>();
		currentSubdomains = new HashMap<String, String>();
	}
	
	/**
	 * Log the user in to allow further actions
	 * @param domainName The name of the domain that user wants to log in
	 * @param password The password.
	 * @throws UnsupportedEncodingException 
	 * @throws NoSuchAlgorithmException 
	 */
	public void login(String domainName, String password) {
		String encrytpedPassword = encryptor.encrypt(password);
		if (isAllowedInDomain(domainName) && encrytpedPassword.equals(getPassword())) {
			liveDomains.add(domainName);
		}
	}
	
	/**
	 * Logs this user out of a domain
	 */
	public void logout(String domainName) {
		if (liveDomains.contains(domainName)) {
			liveDomains.remove(domainName);
		}
	}
	
	/**
	 * Get the login status for the user (ie if they are logged in or not)
	 * for a specific domain
	 * @return True if they are logged in, false otherwise.
	 */
	public boolean isLoggedIn(String domainName) {
		return liveDomains.contains(domainName);
	}
	
	/**
	 * Make sure a student model exists for the given domain
	 * @param domainName The domain (name)
	 */
	public void initModel(String domainName) {
		if(!models.containsKey(domainName) || models.get(domainName)==null) {
			models.put(domainName, StudentModel.emptyModelFactory(getUserName()));
		}
	}
	
	/**
	 * Get the userName of this user.
	 * @return A copy of userName
	 */
	public String getUserName() {
		return new String(userName);
	}
	
	/**
	 * Get the encrypted password of this user.
	 * @return A copy of password
	 */
	public String getPassword() {
		return new String(this.password);
	}
	
	/**
	 * Validate the username which only allows alphabets, numbers and underscore
	 * @param username The student username that needs to be validated
	 * @return 1 if there are any illegal characters; 2 if the username is shorter
	 * than lenght of 2; 0 if the username is correct
	 */
	public static int checkUsername(String username) {
		Pattern p = Pattern.compile("[a-zA-Z_0-9]*");
		Matcher m = p.matcher(username);
		if(!m.matches()){
			return 1;
		}
		if (username.length() < 3) {
			return 2;
		}
		return 0;
	}
	
	/**
	 * Check the password is longer than 2
	 * @param password The student's password that needs to be checked
	 * @return 0 if the password is ok, 2 if the length of the password is shorter than 2
	 */
	public static int checkPassword(String password) {	
		if (password.length() < 3) {
			return 2;
		}
		return 0;
	}
	
	/**
	 * Construct a new Student Object. Encrypts the given password
	 * @param userName The username of the student
	 * @param password The plain text password of the student
	 * @param allowedDomains A set of domain names the student is allowed to access
	 * @return The newly constructed Student object.
	 * @throws IllegalUserNameException If the username is Illegal
	 */
	public static Student newStudentFactory(String userName, String password, Set<String> allowedDomains) 
	throws IllegalUserNameException {
		return new Student(userName, Encryptor.getEncryptor().encrypt(password), allowedDomains);
	}
	
	/**
	 * Change the Student password
	 * @param newPassword The new password that the user wants to change to
	 * @param oldPassword The existing password of a user
	 * @throws InvalidPasswordException If the old password is wrong
	 */
	public void changePassword(String oldPassword, String newPassword) throws InvalidPasswordException {
		String encrytpedPassword = encryptor.encrypt(oldPassword);
		if (!encrytpedPassword.equals(getPassword())) 
			throw new InvalidPasswordException("You must supply the old password!");
		this.password = encryptor.encrypt(newPassword);
		String pathName = Settings.getInstance().getStudentPath();      
        StudentWriter smX = new StudentWriter(this, pathName);	
        smX.changePassword();
	}
	
	@Override
	public boolean equals(Object object1) {
		if (this == object1)
			return true;
		if (object1 == null)
			return false;
		if (getClass() != object1.getClass())
			return false;
		final Student other = (Student) object1;
		
		if (userName == null) {
			if (other.userName != null)
				return false;
		} else if (!userName.equals(other.userName))
			return false;
		return true;
	}

	/**
	 * Standard compareTo override, based on the userName
	 * @param object the Object to be compared
	 * @return a negative integer, zero, or a positive integer as this object
	 * is less than, equal to, or greater than the specified object.
	 * @throws ClassCastException
	 */
	public int compareTo(Object object) throws ClassCastException {
		if (object instanceof Student) {
			Student other = (Student) object;
			return this.getUserName().compareTo(other.getUserName());
		} else {
			throw new ClassCastException();
		}
	}
	
	@Override
	public int hashCode() {
		int result = userName.hashCode();
		return result;
	}
	
	/**
	 * Checks whether a Student has access to a domain or not.
	 * @param domainName The name of the domain we want to check
	 * @return True if the student has access to the domain, otherwise false
	 */
	public boolean isAllowedInDomain(String domainName) {
		if(allowedDomains.contains(domainName)) return true;
		return false;
	}
	
	/**
	 * Returns a complete unmodifiable Set of all the domains a student can log into
	 * @return a Set of all domains a student can log into
	 */
	public Set<String> getAllowedDomains() {
		return Collections.unmodifiableSet(allowedDomains);
	}
	
	/**
	 * Adds a new domain to the set of domains a student can log into.
	 * Makes a new StudentModel for the domain and writes it to XML
	 * If the domain does not allow student creation, then a AddUserDeniedException
	 * @param domainName The domain name that the students can log into
	 * @throws AddUserDeniedException 
	 * @throws StudentModelAlreadyExistsException 
	 * @throws UserAlreadyExistsException 
	 */
	public void addAllowedDomain(String domainName) throws StudentModelAlreadyExistsException {
		if(getStudentModel(domainName)!=null)
			throw new StudentModelAlreadyExistsException("Student "+getUserName()+" already has a model for domain "+domainName);
		createStudentModel(domainName);
		addDomainToGlobalStudentInfo(domainName);
		allowedDomains.add(domainName);
	}
	
	/**
	 * Add a student model to this students models. Gets called when loading students from XML
	 * @param domainName The name of the domain this model is for
	 * @param model The StudentModel object
	 */
	public void addStudentModel(String domainName, StudentModel model) {
		models.put(domainName, model);
	}
	
	/**
	 * Creates a new (empty) StudentModel and adds it to this students models.
	 * @param domainName
	 */
	public void createStudentModel(String domainName) {
		StudentModel model = StudentModel.emptyModelFactory(getUserName());
		models.put(domainName, model);
	}
	
	private void deleteStudentModel(String domainName) {
		models.remove(domainName);
	}
	
	private StudentModel getStudentModel(String domainName) {
		return models.get(domainName);
	}
	
	/*
	 * Adds a new domain to the set of allowed domains in the students.xml file
	 */
	private void addDomainToGlobalStudentInfo(String domainName) {
		StudentWriter studWriter = new StudentWriter(this, Settings.getInstance().getStudentPath());
		studWriter.addAllowedDomain(domainName);
	}
	
	/**
	 * Only Removes a domain from the students.xml file,
	 * Does not remove the student model, if you want to remove
	 * the student model as well JUST call removeUser(Student student, Domain domain)
	 * @param domainName The name of the domain that the student needs to be removed
	 */
	private void removeAllowedDomain(String domainName) {
		allowedDomains.remove(domainName);
		String pathname = Settings.getInstance().getStudentPath();
		StudentWriter writer = new StudentWriter(this,pathname);
		writer.removeDomain(domainName);
	}
	
	/**
	 * Returns a complete Set of all the domains a student is currently logged into
	 * @return liveDomains A Set of all domains a student is currently logged into
	 */
	public Set<String> getLiveDomains() {
		return Collections.unmodifiableSet(liveDomains);
	}

	/**
	 * Loads all the Students from xml, and returns a set of those (loaded) Students.
	 * @return Set<Student> The set of all the studnets in the system
	 * @throws FileNotFoundException
	 */
	public static Set<Student> loadAllStudents() throws FileNotFoundException{
		String studentPath = Settings.getInstance().getStudentPath();
		String rootPath = Settings.getInstance().getRootPath();
		String modelsFileName = Settings.getInstance().getStudentModelsFileName();
		StudentParser parser = new StudentParser(studentPath, rootPath, modelsFileName);
		return parser.processTree();
	}
	
	/**
	 * Saves the Student to xml, creates a new StudentModel for the domain, and also writes
	 * that to XML. This should be used only when a new Student is being created.
	 * @param student The Student to be persisted
	 * @throws AddUserDeniedException
	 * @throws StudentModelAlreadyExistsException 
	 * @throws UserAlreadyExistsException
	 */
	public static void persistStudent(Student student) {		
		/* Writes Student object to xml*/
		String pathname = Settings.getInstance().getStudentPath();
		StudentWriter writer = new StudentWriter(student, pathname);	
		writer.addStudent();		
	}
	
	/**
	 * Removes access to a specific Domain, for a Student.
	 * Note: this does not delete a student from the system, only from a certain Domain
	 * @param student The student that needs to be removed
	 * @param domainName the domain which you want a student to have revoked access on.
	 * @throws UserNotFoundException if a Students student model is not found in domain
	 */
	public static void removeUser(Student student, String domainName) throws UserNotFoundException {
		if(student.getAllowedDomains().contains(domainName)) {
			student.removeAllowedDomain(domainName);
		} else {
			Logger.addEntry("Tried to remove a user that didn't exist");
		}
		
		student.deleteStudentModel(domainName);		
		Settings settings = Settings.getInstance();
		StudentWriter globalStudentInfo = new StudentWriter(student, settings.getStudentPath());
		globalStudentInfo.removeDomain(domainName);
	}
		
	/**
	 * Deletes students global information -NOT domain information.
	 * To delete student models, use removeUser(Student, Domain).
	 * @param student the student which is to be deleted.
	 * @throws UserNotFoundException
	 */
	public static void deleteUser(Student student) throws UserNotFoundException {
		String pathname = Settings.getInstance().getStudentPath();
		StudentWriter writer = new StudentWriter(student, pathname);
		writer.removeStudent();
		student = null;
	}

	/**
	 * Set the current working subdomain for a given domain
	 * @param domainName The domain name
	 * @param currentSubdomain The name of the current subdomain
	 */
	public void setCurrentSubdomain(String domainName, String currentSubdomain) {
		currentSubdomains.put(domainName, currentSubdomain);
		StudentModel model = getStudentModel(domainName);
		model.changedSubdomain(currentSubdomain);
	}

	/**
	 * Get the name of the current subdomain of a particular domain
	 * @param domainName The domain name
	 * @return The name of the current subdomain for the given domain.
	 */
	public String getCurrentSubdomain(String domainName) {
		return currentSubdomains.get(domainName);
	}
	
	/**
	 * Gets the next problem the student hasnt solved in the current subdomain of a given domain
	 * @param domainName The name of the domain
	 * @param problems A collection of problems to choose the next from
	 * @return The next problem
	 */
	public Problem getNextProblem(String domainName, SortedMap<Integer, Problem> problems) {
		StudentModel model = getStudentModel(domainName);
		return model.getNextProblem(problems, getCurrentSubdomain(domainName));
	}
	
	/**
	 * Returns a problem, given a problemID, for a domain
	 * @param domainName The name of the domain
	 * @param problems The problems to select from
	 * @param problemNumber The ID of they wanted problem
	 * @return The problem
	 */
	public Problem getNextProblem(String domainName, SortedMap<Integer, Problem> problems, int problemNumber) {
		StudentModel model = getStudentModel(domainName);
		return model.getNextProblem(problems, problemNumber);
	}

	/**
	 * Get the status of a given problem, from a given domain.
	 * @param domainName The name of the domain
	 * @param problemID The ID of the problem
	 * @return The status eg "attempted" etc
	 */
	public String getProblemStatus(String domainName, int problemID) {
		StudentModel model = getStudentModel(domainName);
		return model.getProblemStatus(getCurrentSubdomain(domainName), problemID);
	}
	
	/**
	 * Get the ID of the problem the student is currently working on for a given domain
	 * @param domainName The name of the domain
	 * @return The ID of the current problem
	 */
	public Integer getCurrentProblemID(String domainName) {
		StudentModel model = getStudentModel(domainName);
		return model.getCurrentProblemID();
	}
	
	/**
	 * Set the status for a given problem.
	 * @param domainName The domain name
	 * @param subdomainName The subdomain name the problem belongs to
	 * @param problemID The ID of the problem
	 * @param status The status
	 */
	public void setProblemStatus(String domainName, String subdomainName, int problemID, String status) {
		StudentModel model = getStudentModel(domainName);
		model.setProblemStatus(subdomainName, problemID, status);
	}
	
	/**
	 * Set this students rating for a given domain.
	 * @param domainName The domain name
	 * @param rating The new rating
	 */
	public void setRating(String domainName, double rating) {
		StudentModel model = getStudentModel(domainName);
		model.setRating(rating);
	}
	
	/**
	 * Get the rating of this student for a given domain.
	 * @param domainName The domain name
	 * @return The rating for the given domain
	 */
	public double getRating(String domainName) {
		StudentModel model = getStudentModel(domainName);
		return model.getRating();
	}
	
	/**
	 * Increment this students rating in the given domain by a certain amount
	 * @param domainName The domain name
	 * @param increment The amount to increase the rating by
	 */
	public void incrementRating(String domainName, double increment) {
		StudentModel model = getStudentModel(domainName);
		model.incrementRating(increment);
	}

	/**
	 * Set the status of the current problem for a domain
	 * @param domainName The domain name
	 * @param status The status for the current problem
	 */
	public void setCurrentProblemStatus(String domainName, String status) {
		StudentModel model = getStudentModel(domainName);
		String subdomainName = currentSubdomains.get(domainName);
		model.setCurrentProblemStatus(subdomainName, status);
		
	}

	/**
	 * Add a problem to the collection of solved problems for a subdomain
	 * @param domainName The domain name
	 * @param subdomainName The name of the subdomain the problem belongs to
	 * @param problemID The ID of the solved problem
	 */
	public void addSolvedProblem(String domainName, String subdomainName, Integer problemID) {
		StudentModel model = getStudentModel(domainName);
		model.addSolvedProblem(problemID, subdomainName);
	}

	/**
	 * Add some violated constraints to this students model for the given domain
	 * @param domainName The domain name
	 * @param violatedConstraints A collection of constraints
	 */
	public void addViolatedConstraints(String domainName, Collection<String> violatedConstraints) {
		StudentModel model = getStudentModel(domainName);
		model.addViolatedConstraints(violatedConstraints);
	}

	/**
	 * Add some satisfied constraints to this students model for the given domain
	 * @param domainName the domain name
	 * @param satisfiedConstraints A collection of constraints
	 */
	public void addSatisfiedConstraints(String domainName, Collection<String> satisfiedConstraints) {
		StudentModel model = getStudentModel(domainName);
		model.addSatisfiedConstraints(satisfiedConstraints);
	}

	/**
	 * Get the IDs of the problems solved for a given subdomain
	 * @param domainName The domain name
	 * @param subdomainName The subdomain name
	 * @return The IDs of all the solved problems in the subdomain
	 */
	public Set<Integer> getSolvedProblems(String domainName, String subdomainName) {
		StudentModel model = getStudentModel(domainName);
		return model.getSolvedProblems(subdomainName);
	}

	/**
	 * Get the collection of constraints satisfied in the given domain
	 * @param domainName The domain name
	 * @return All the satisfied constraints for the given domain
	 */
	public Collection<String> getSatisfiedConstraints(String domainName) {
		StudentModel model = getStudentModel(domainName);
		return model.getSatisfiedConstraints();
	}

	/**
	 * Get the collection of constraints violated in the given domain
	 * @param domainName The domain name
	 * @return All the violated constraints for the given domain
	 */
	public Collection<String> getViolatedConstraints(String domainName) {
		StudentModel model = getStudentModel(domainName);
		return model.getViolatedConstraints();
	}

	/**
	 * Save the model for a given domain
	 * @param domainName The domain name
	 */
	public void saveStudentModel(String domainName) {
		StudentModel model = getStudentModel(domainName);
		model.save(domainName);
	}
}
