package jitas.core;

import jitas.exception.*;
import jitas.util.*;

import java.io.FileNotFoundException;
import java.io.IOException;
import java.rmi.RemoteException;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;
import java.util.SortedMap;
import java.util.TreeSet;

import jess.JessException;

/**
 * The main class for the JITAS class library (core classes). All other
 * applications will use this class to interface with JITAS. This class
 * provides all the necessary functionality for a higher level application
 * to act as an Intelligent Tutoring System. To use this class it needs to
 * be imported by the calling class, which can then directly call methods on
 * this class.
 * 
 * @author	DeathMarch (c314g2)
 * @version	24 August 2007
 */
public class Tutor implements TutorInterface {
	
	private Set<Student> students;
	private Map<String, Domain> domains;
	
	/**
	 * The default constructor for the tutor class.
	 * This method should be called whenever an application wants to create a
	 * new Tutor. All subsequent method calls should be made on that Tutor
	 * object.
	 * eg. In the higher level application:
	 * 		Tutor t = new Tutor();
	 * 		...
	 * @throws IOException 
	 */
	public Tutor() throws IOException {
		students = new HashSet<Student>();
		initDomains();
		try {
			students = Student.loadAllStudents();
		} catch (FileNotFoundException e) {
			System.err.println(e);
			throw e;
		}
	}
	
	/**
	 * Constructor that takes in a file path for the domains.xml file
	 * @param domainsFilePath The path to an xml file containing information
	 * 	about all the domains. The path must include the filename.
	 * @throws IOException 
	 */
	public Tutor (String domainsFilePath) throws IOException {
		students = new HashSet<Student>();
		initDomains(domainsFilePath);
		try {
			students = Student.loadAllStudents();
		} catch (FileNotFoundException e) {
			System.err.println(e);
			throw e;
		}
	}
	
	public void login(String userName, String password, String domainName)
	throws DomainNotFoundException, DomainNotLoadedException, 
	InvalidLoginException, RemoteException, UserNotFoundException {
		
		checkDomainIsLoaded(domainName);
		
		Student student = getStudentByName(userName);
		student.login(domainName, password);
		
		if (!student.isLoggedIn(domainName)) {
			throw new InvalidLoginException("Tutor.login: Invalid Login.");
		}
		
		student.initModel(domainName);
	}
	
	/*Check whether the domain is loaded*/
	private void checkDomainIsLoaded(String domainName) 
	throws DomainNotFoundException, DomainNotLoadedException {
		getDomain(domainName); //throws exceptions
	}
	
	/* Returns the desired Student Object. */
	private Student getStudentByName(String userName)
	throws UserNotFoundException {		
		for(Student s : students) {
			if(s.getUserName().equals(userName)) return s;
		}
		throw new UserNotFoundException("Tutor.getStudentByName: user: " + userName + " does not exist!");
	}
	
	public void logout(String userName, String domainName)
	throws DomainNotFoundException, DomainNotLoadedException, UserNotFoundException,
	UserNotLoggedInException, RemoteException {
		checkDomainIsLoaded(domainName);
		if (getLoginStatus(userName, domainName)) {
			getStudentByName(userName).logout(domainName);
		} else {
			throw new UserNotLoggedInException("Tutor.logout: User not logged in.");
		}
	}
	
	public boolean getLoginStatus(String userName, String domainName)
	throws UserNotFoundException, RemoteException {
		return getStudentByName(userName).isLoggedIn(domainName);
	}
	
	/* Throws an UserNotLoggedInException if the user is not logged in. */
	private void checkLoggedIn(String userName, String domainName, String methodName)
	throws UserNotFoundException, UserNotLoggedInException, RemoteException {
		if (!getLoginStatus(userName, domainName)) {
			throw new UserNotLoggedInException(methodName + " You must be logged in to do that!");
		}
	}
	
	/* Implements addUser in TutorInterface */
	public void addUser(String userName, String password, String domainName)
	throws AddUserDeniedException, DomainNotFoundException, DomainNotLoadedException,
	IllegalUserNameException, StudentModelAlreadyExistsException, RemoteException {
		addUser(userName, password, domainName, false);
	}
	
	/* Note: this method allows any user to attempt to add a new domain
	 * to any other user.*/
	public void addUser(String userName, String password, String domainName, boolean bypassSecurity)
	throws AddUserDeniedException, DomainNotFoundException, DomainNotLoadedException,
	IllegalUserNameException, StudentModelAlreadyExistsException, RemoteException {
		
		checkDomainIsLoaded(domainName);
		Domain domain = getDomain(domainName);
		
		if (!domain.userCreationAllowed() && !bypassSecurity) 
			throw new AddUserDeniedException("Domain "+domainName+" does not allow you to add user "+userName);
		
		Student student;
		try {
			student = getStudentByName(userName);
			student.addAllowedDomain(domainName);
		} catch (UserNotFoundException e) {
			Set<String> allowedDomains = new HashSet<String>();
			allowedDomains.add(domainName);			
			student = Student.newStudentFactory(userName, password, allowedDomains);
			student.createStudentModel(domainName);
			students.add(student);
			Student.persistStudent(student);
		}
	}
	
	public void removeUser(String userName, String domainName)
	throws DomainNotFoundException, DomainNotLoadedException, UserNotFoundException,
	UserNotLoggedInException, RemoteException {
		/* Remove a user from ONE domain */
		checkDomainIsLoaded(domainName);
		checkLoggedIn(userName,domainName,"Tutor.removeUser");
		logout(userName, domainName);
		Student.removeUser(getStudentByName(userName), domainName);
	}
	
	public void deleteUser(String userName)
	throws DomainNotFoundException, DomainNotLoadedException, UserNotFoundException,
	UserNotLoggedInException, RemoteException {
		Student student = getStudentByName(userName);
		students.remove(student);
		Set<String> set = new TreeSet<String>();
		
		for (String d : student.getLiveDomains()) {
			set.add(d);
		}
		
		set = new TreeSet<String>();
		
		for (String d : student.getAllowedDomains()) {
			set.add(d);
		}
		
		for (String d : set) {
			Student.removeUser(student, d);
		}
		Student.deleteUser(student);
	}
	
	public void addAllowedDomain(String userName, String domainName)
	throws AddUserDeniedException, DomainNotFoundException, DomainNotLoadedException,
	StudentModelAlreadyExistsException, UserNotFoundException, RemoteException {
		addAllowedDomain(userName, domainName, false);
	}
	
	public void addAllowedDomain(String userName, String domainName, boolean bypassSecurity)
	throws AddUserDeniedException, DomainNotFoundException, DomainNotLoadedException,
	StudentModelAlreadyExistsException, UserNotFoundException, RemoteException {
		
		checkDomainIsLoaded(domainName);
		
		if(!getAllDomainNames().contains(domainName))
			throw new DomainNotFoundException("Domain +"+domainName+" not found");
		
		Domain domain = getDomain(domainName);
		if(!domain.userCreationAllowed())
			throw new AddUserDeniedException("You cannot add users to domain "+ domainName);
		
		Student student = getStudentByName(userName);
		student.addAllowedDomain(domainName);
	}
	
	public void changePassword(String userName, String oldPassword, String newPassword)
	throws InvalidPasswordException, UserNotFoundException, RemoteException {
		Student student = getStudentByName(userName);
		student.changePassword(oldPassword, newPassword);
	}
	
	public Set<String> getAllStudentUsernames() {
		Set<String> names = new HashSet<String>();
		for(Student s : students) {
			names.add(s.getUserName());
		}
		return names;
	}
	
	public Set<String> getAllStudentUsernames(String filter) {
		Set<String> names = new HashSet<String>();
		for(Student s : students) {
			if(s.isAllowedInDomain(filter)) names.add(s.getUserName());
		}
		return names;
	}
	
	public boolean isStudentAllowedInDomain(String userName, String domainName) throws UserNotFoundException {
		Student student = getStudentByName(userName);
		return student.isAllowedInDomain(domainName);
	}
	

	public boolean isUserCreationAllowed(String domainName) throws DomainNotFoundException, DomainNotLoadedException {
		Domain domain = getDomain(domainName);
		return domain.userCreationAllowed();
	}
	
	public void refreshStudentList() throws FileNotFoundException {
		students = Student.loadAllStudents();
	}
	
	private void initDomains() {
		initialiseDomains();
		Settings s = Settings.getInstance();
		s.setRootPath("domains/");
	}
	
	private void initDomains(String domainsFilePath) {
		Settings s = Settings.getInstance();
		s.setRootPath(domainsFilePath);
		initialiseDomains();
	}
	
	private void initialiseDomains() {
		domains = new HashMap<String, Domain>();
		try {
			Set<String> domainNames = Domain.getDomainNames();
			for(String name : domainNames) {
				domains.put(name, null);
			}
		} catch (IOException e) {
			System.err.println("FATAL ERROR! Tutor.initDomains: "+e.getMessage());
		}
	}	
	
	/**
	 *  Returns the names of all the domains this tutor knows about (loaded or otherwise)
	 *  @return Set of domain names 
	 */
	public Set<String> getAllDomainNames()
	throws RemoteException {
		return domains.keySet();
	} 
	
	/**
	 * This gets the long domain names for all loaded domains
	 * @return Map<domainName, longDomainName>
	 */
	public Map<String, String> getAllLongDomainNames()
	throws RemoteException {
		Map<String, String> longNames = new HashMap<String, String>();
		for(String domainName : domains.keySet()) {
			try {
				Domain domain = getDomain(domainName);
				longNames.put(domain.getName(), domain.getLongName());
			} catch (DomainNotFoundException e) { //ignore
			} catch (DomainNotLoadedException e) { //ignore
			}
		}
		return longNames;		
	}
	
	public Set<String> getSubdomains(String domainName)
	throws DomainNotFoundException, DomainNotLoadedException, RemoteException {
		/* Note: the domain must be loaded to get the subdomain set */
		Domain domain = getDomain(domainName);
		return domain.getSubdomainNames(); 
	}
	
	/**
	 * Load the domain from XML
	 */
	public void loadDomain(String domainName)
	throws DomainNotFoundException, RemoteException {
		/* we need to load it from xml, if the domain is not found,
		 *  the following code will throw an exception            
		 */	
		Domain domain = Domain.loadDomain(domainName);
		
		if(!domains.containsKey(domainName)) {
			Domain.addDomain(domain);
		}
		domains.put(domainName, domain);
		try {
			domain.loadSubdomains();
			domain.readRules();
		} catch (ProblemsNotFoundException e) {
			System.err.println("WARNING! " + e.getMessage());
		} catch (JessException e) {
			System.err.println("WARNING! while loading rules for domain "+domainName+": " + e.getMessage());
		}
	}
	
	/* Returns the desired Domain object. */
	private Domain getDomain(String domainName)
	throws DomainNotFoundException, DomainNotLoadedException {
		if(!domains.containsKey(domainName)) 
			throw new DomainNotFoundException("Domain \"" + domainName + "\" does not exist.");
		if(domains.get(domainName)==null)
			throw new DomainNotLoadedException("Domain \"" + domainName + "\" exists, but has not been loaded");

		return domains.get(domainName);
	}

	public void selectSubdomain(String userName, String subdomainName, String domainName)
	throws DomainNotFoundException, DomainNotLoadedException, ProblemsNotFoundException,
	SubdomainNotFoundException, UserNotFoundException, UserNotLoggedInException,
	RemoteException {
		checkDomainIsLoaded(domainName);
		checkLoggedIn(userName,domainName,"Tutor.selectSubdomain");
		Student student = getStudentByName(userName);
		student.setCurrentSubdomain(domainName, subdomainName);		
	}
	
	public Problem getNextProblem(int problemNumber, String userName, String domainName)
	throws DomainNotFoundException, DomainNotLoadedException, InvalidProblemException,
	UserNotFoundException, UserNotLoggedInException, RemoteException {
		return getNextProblemHelper(problemNumber, userName, domainName);
	}
		
	public Problem getNextProblem(String userName, String domainName)
	throws DomainNotFoundException, DomainNotLoadedException, InvalidProblemException,
	UserNotFoundException, UserNotLoggedInException, RemoteException {
		return getNextProblemHelper(-1, userName, domainName);
	}
	
	/* If ProblemNumber is less than 0, means that the problem number is not used to 
	retrieve the next problem. */
	private Problem getNextProblemHelper(int problemNumber, String userName, String domainName)
	throws DomainNotFoundException, DomainNotLoadedException, InvalidProblemException,
	UserNotFoundException, UserNotLoggedInException, RemoteException {
		Domain domain = getDomain(domainName);
		checkLoggedIn(userName,domainName, "Tutor.getNextProblemHelper");
		
		Student student = getStudentByName(userName);
		String subdomainName = student.getCurrentSubdomain(domainName);
		SortedMap<Integer, Problem> problems;
		try {
			problems = domain.getProblems(subdomainName);
		} catch (SubdomainNotFoundException e) {
			System.err.println("FATAL: subdomain "+subdomainName+" not found"); 
			throw new InvalidProblemException("Tutor.getNextProblemHelper: Cannot get problem");
		}
		
		Problem p;
		if(problemNumber < 0){
			p = student.getNextProblem(domainName, problems);
		} else {
			p = student.getNextProblem(domainName, problems, problemNumber);
		}
		if (p == null) {
			throw new InvalidProblemException("Tutor.getNextProblemHelper: Cannot get problem");
		}
		return p;
	}
	
	public void setCustomProblemSelectionAlgorithm(String domainName, String fileName) 
	throws DomainNotFoundException, DomainNotLoadedException {
		checkDomainIsLoaded(domainName);
		Domain domain = getDomain(domainName);
		domain.loadSpecificAlgorithm(fileName);		
	}
	
	public void setCustomFeedback(String domainName, String fileName) 
	throws DomainNotFoundException, DomainNotLoadedException {
		checkDomainIsLoaded(domainName);
		Domain domain = getDomain(domainName);		
		domain.loadSpecificFeedback(fileName);		
	}
	
	public Problem getProblem(String userName, String domainName)
	throws DomainNotFoundException, DomainNotLoadedException, InstantiationException,
	IllegalAccessException, RemoteException, UserNotFoundException, InvalidProblemException {
		
		Problem p = null;
		SortedMap<Integer, Problem> problems;
		
		Domain domain = getDomain(domainName);		
		domain.getSelectionAlgorithm();
		
		Student student = getStudentByName(userName);
		String subdomainName = student.getCurrentSubdomain(domainName);
		
		try {
			problems = domain.getProblems(subdomainName);
		} catch  (SubdomainNotFoundException e) {
			System.err.println("FATAL: subdomain "+subdomainName+" not found"); 
			throw new InvalidProblemException("Tutor.getNextProblemHelper: Cannot get problem");
		}
		//p = student.getNextProblem(domainName, problems, domain.getSelectionAlgorithm());
		
		p = domain.getNextProblem(student, domain, problems, subdomainName);
		
		
		return p;
	}
	
	public SortedMap<Integer, Problem> getAllProblems(String userName,String domainName)
	throws DomainNotFoundException, DomainNotLoadedException, UserNotFoundException,
	UserNotLoggedInException, RemoteException {
		Domain domain = getDomain(domainName);
		checkLoggedIn(userName,domainName,"Tutor.getAllProblems");
		Student student = getStudentByName(userName);
		String subdomainName = student.getCurrentSubdomain(domainName);
		
		try {
			return domain.getProblems(subdomainName);
		} catch (SubdomainNotFoundException e) {
			throw new DomainNotFoundException("Subdomain "+subdomainName+" not found");
		}
	}
	
	public String getProblemStatus(String userName, String domainName, int problemID) 
	throws UserNotFoundException, UserNotLoggedInException, RemoteException, DomainNotFoundException, DomainNotLoadedException {
		checkLoggedIn(userName,domainName,"Tutor.getProblemStatus");
		Student student = getStudentByName(userName);
		return student.getProblemStatus(domainName, problemID);
	}
	
	public Feedback submitSolution(Solution studentSolution, String userName, String domainName)
	throws DomainNotFoundException, DomainNotLoadedException, InvalidSolutionException,
	StudentModelNotFoundException, UserNotFoundException, UserNotLoggedInException,
	JessException, RemoteException{
		checkLoggedIn(userName,domainName,"Tutor.submitSolution");
		Domain domain = getDomain(domainName);
		Student student = getStudentByName(userName);
		String subdomainName = student.getCurrentSubdomain(domainName);
		return SolutionEvaluator.evaluateSolution(studentSolution, student, domain ,subdomainName, true);
	}
}
