package jitas.util;

/**
 * Settings class is mainly used to configure the paths to different components
 * of a Tutor.
 * 
 * Currently it uses a singleton-style getSettings method for access by other 
 * classes.  
 * 
 * @author	DeathMarch (c314g2)
 * @version	10 July 2007
 */
public class Settings {

	static Settings settings = null;	
	static final String allDomainsFileName = "domains.xml";
	private String rootPath = "domains/";
	private String studentFileName = "students.xml";
	private String problemSelectionAlgorithmFile =  "CustomProblemSelectionAlgorithm.jar";
	private String feedbackObjectFile = "Feedback.jar";
	private String JESSParserFile = "SpecificJESSParser.jar";
	private String studentModelsFileName =  "/student-models.xml"; //used to find the student models


	/**
	 * The default constructor for Setting object
	 */
	private Settings() {}

	/**
	 * Get an instance of a Settings object
	 * @return An instance of a Settings object
	 */
	public static Settings getInstance() {
		if (settings == null) {
			settings = new Settings();
		}
		return settings;
	}
	
	/**
	 * Get the name of the file that contains all the domain names
	 * @return The file name
	 */
	public String getAllDomainsFileName() {
		return allDomainsFileName;
	}
	
	/**
	 * Set the root path of all the docs related to the domain
	 * @param s The root path
	 */
	public void setRootPath(String s) {
		rootPath = s;
	}
	
	/**
	 * Get the root path of where the documents are
	 * @return The root path
	 */
	public String getRootPath() {
		return new String(rootPath);
	}
	
	/**
	 * Get the full path to AllDomains xml file
	 * @return The full path to the file which contains all the domain names
	 */
	public String getFullPathToAllDomains() {
		return rootPath + allDomainsFileName;
	}
	
	/**
	 * Get the full path to a semantic constraint file of a specific domain 
	 * @param nameOfDomain The name of the domain which requires its constraint path
	 * @return The path to a specific domain constraint file
	 */
	public String getSemanticConstraintsPath(String nameOfDomain) {
		return rootPath + nameOfDomain + "/semantic-constraints.clp";
	}

	/**
	 * Get the full path to a syntax constraint file of a specific domain 
	 * @param nameOfDomain The name of the domain which requires its constraint path
	 * @return The path to a specific domain constraint file
	 */
	public String getSyntaxConstraintsPath(String nameOfDomain) {
		return rootPath + nameOfDomain + "/syntax-constraints.clp";
	}	
	
	/**
	 * Get the full path to a semantic constraint file of a specific domain and subdomain
	 * @param nameOfDomain The name of the domain which requires its constraint path
	 * @param subDomainName The name of the subdomain
	 * @return The path to a specific domain constraint file
	 */
	public String getSemanticConstraintsPath(String nameOfDomain, String subDomainName) {
		return rootPath + nameOfDomain + "/" + subDomainName + "-" + "semantic-constraints.clp";
	}

	/**
	 * Get the full path to a syntax constraint file of a specific domain and subdomain
	 * @param nameOfDomain The name of the domain which requires its constraint path
	 * @param subDomainName The name of the subdomain
	 * @return The path to a specific domain constraint file
	 */
	public String getSyntaxConstraintsPath(String nameOfDomain, String subDomainName) {
		return rootPath + nameOfDomain + "/" + subDomainName + "-" + "syntax-constraints.clp";
	}
	
	/**
	 * Gets the global student details file
	 * @return the name of the global student details file
	 */
	public String getStudentFileName() {
		return studentFileName;
	}

	/**
	 * Sets the global student details file
	 * @param studentFileName the name of the global student details file
	 */
	public void setStudentFileName(String studentFileName) {
		this.studentFileName = studentFileName;
	}
	
	/**
	 * Gets the global student details file path.
	 * @return the path to the global student details file
	 */
	public String getStudentPath() {
		return rootPath + studentFileName;
	}
	
	/**
	 * Get the path to the specific domain Problem Selection Algorithm file
	 * @param nameOfDomain The name of the domain
	 * @return the path to the specific domain Problem Selection Algorithm file
	 */
	public String getAlgorithmFileName(String nameOfDomain) {
		return rootPath + nameOfDomain + "/" + problemSelectionAlgorithmFile;
	}

	/**
	 * Get the name of the jar containing the custom feedback object
	 * @param nameOfDomain The name of the domain
	 * @return The name of the jar containing the custom feedback object
	 */
	public String getFeedbackObjectFile(String nameOfDomain) {
		return rootPath + nameOfDomain + "/" + feedbackObjectFile;
	}

	/**
	 * Get the name of the jar containing the custom JESSParser object
	 * @param nameOfDomain The name of the domain
	 * @return The name of the jar containing the custom JESSParser object
	 */
	public String getJESSParserFile(String nameOfDomain) {
		return rootPath + nameOfDomain + "/" + JESSParserFile;
	}

	/**
	 * Get the file name for the student models
	 * @return Student models file name
	 */
	public String getStudentModelsFileName() {
		return studentModelsFileName;
	}

	public String getStudentModelPath(String domainName) {
		return rootPath + domainName + studentModelsFileName;
	}
}
