package jitas.xml.writer;

import java.util.Iterator;
import jitas.core.StudentModel;
import jitas.exception.UserNotFoundException;
import jitas.util.Logger;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.w3c.dom.Text;

/**
 * A class that provides functions for users to write to student-models.xml.
 * 
 * Future enhancements could include bulk processing of persisting student models
 * which would mean that the student models xml file wouldn't need to be opened and
 * closed multiple times.
 * 
 * @author	DeathMarch (c314g2)
 * @version	21 August 2007
 */
public class StudentModelWriter extends XMLWriter{
	
	private StudentModel sm = null;
	private Document parsedDoc = null;
	private String pathName = "";
	
	/**
	 * Constructor. Pass the parameters to the superclass to load
	 * the XML file.
	 */
	public StudentModelWriter(StudentModel sm, String pathName){
		this.sm = sm;
		this.pathName = pathName;
		parsedDoc = super.loadXML(pathName);
	}
	
	/**
	 * Add a student model to the student-model XML file.
	 */
	public void addStudentModel(){
		/*Make sure that the File existed, and returned a valid document */
		if(parsedDoc == null) {
			Logger.addEntry("Could not add a Student Model in StudentModelWriter, as parsedDoc is null");
			return;
		}
		
		
		
		/* Create a student model*/
        Element studentModel = parsedDoc.createElement("student-model");
        
        /* Create username element and attach it to studentModel */
		Element usernameEle = parsedDoc.createElement("username");
		Text usernameText = parsedDoc.createTextNode(sm.getUserName());
		usernameEle.appendChild(usernameText);
		
		Element problemsStatusEle = parsedDoc.createElement("problems-status");
		Text problemsStatusText = parsedDoc.createTextNode(sm.getAllProblemsStatus().toString());
		problemsStatusEle.appendChild(problemsStatusText);
		
		Element solvedProblemEle = parsedDoc.createElement("solved-problems");
		Text solvedProblemsText = parsedDoc.createTextNode(sm.getSolvedProblems().toString());
		solvedProblemEle.appendChild(solvedProblemsText);
		
		Element abilityRating = parsedDoc.createElement("rating");
		Text abilityRatingText = parsedDoc.createTextNode(Double.toString(sm.getRating()));
		abilityRating.appendChild(abilityRatingText);
		
		Element constraintsSatisfied = parsedDoc.createElement("constraints-satisfied");
		Iterator<String> it = sm.getSatisfiedConstraints().iterator();
		while(it.hasNext()){
			Element constraint = parsedDoc.createElement("constraint");
			Text constraintText = parsedDoc.createTextNode(it.next());
			constraint.appendChild(constraintText);
			constraintsSatisfied.appendChild(constraint);
		}
		
		Element constraintsViolated = parsedDoc.createElement("constraints-violated");
		Iterator<String> itt = sm.getViolatedConstraints().iterator();
		while(itt.hasNext()){
			Element constraint = parsedDoc.createElement("constraint");
			Text constraintText = parsedDoc.createTextNode(itt.next());
			constraint.appendChild(constraintText);
			constraintsViolated.appendChild(constraint);
		}
		
		
		/*Append the created username, problems-status & solved-problems elements to the student model element*/
		studentModel.appendChild(usernameEle);
		studentModel.appendChild(problemsStatusEle);
		studentModel.appendChild(solvedProblemEle);
		studentModel.appendChild(constraintsSatisfied);
		studentModel.appendChild(constraintsViolated);
		studentModel.appendChild(abilityRating);
		
		if(documentIsEmpty()) {
			parsedDoc.setXmlVersion("1.0");
			Element studentModels = parsedDoc.createElement("student-models");
			studentModels.appendChild(studentModel);
			parsedDoc.appendChild(studentModels);
		} else {
			/*Append the student model element to the root element*/
			parsedDoc.getFirstChild().appendChild(studentModel);
		}
		
		
		/*Write to the XML file*/
		super.writeToXML(parsedDoc,pathName);
	}
	
	/**
	 * Removes a student model from the student-model XML file.
	 */
	public void removeStudentModel() throws UserNotFoundException{
		/*Make sure that the File existed, and returned a valid document */
		if(parsedDoc == null) {
			Logger.addEntry("Could not remove a Student Model in StudentModelWriter, as parsedDoc is null");
			return;
		}
		
		NodeList usernames = parsedDoc.getElementsByTagName("username");
		String userName = sm.getUserName();
		boolean userExists = false;
		
		for (int i = 0; i < usernames.getLength() ; i++) {
			Node n = usernames.item(i);	
			
			if (n.getTextContent().equals(userName)) {
				userExists = true;
				
				Node studentModelNode = n.getParentNode();
				Node parentNode = studentModelNode.getParentNode();
				
				parentNode.removeChild(studentModelNode);
			}
		}
		
		if(!userExists) throw new UserNotFoundException("StudentModelWriter.removeStudentModel"
									+ "could not find the student with username: " + userName);
		
		
		/*Write to the XML file*/
		super.writeToXML(parsedDoc,pathName);
	}	
	
	
	/**
	 * Saves all the current information stored in the Student Model given
	 * to the constructor to the student models xml file.
	 * @throws UserNotFoundException if there is no student model for the student
	 */
	public void persistStudentModel() throws UserNotFoundException {
		try { 
			removeStudentModel();
		} catch(UserNotFoundException exc) {
			throw new UserNotFoundException("StudentModelWriter.persistStudentModel "
					+ "could did not find student with userName: " + sm.getUserName());
		}
		addStudentModel();
	}
	
	/*
	 * If the parsedDoc has no child nodes, it is assumed that
	 * the document is empty.  This method will return true if the
	 * document has no children, otherwise returns false
	 */
	private boolean documentIsEmpty() {
		if(!parsedDoc.hasChildNodes())return true;
		return false;
	}

	
	/*private String print(SortedMap<String, Problem> problems) {
		Set<String> s = new HashSet<String>();
		for(String str : problems.keySet()) {
			s.add(str);
		}
		return s.toString();
	}*/
}
